from typing import Union

import numpy as np
import pandas as pd

FrameOrSeries = Union[pd.DataFrame, pd.Series]

ALL_NAN_TYPE = pd.Int8Dtype()

BOOLEAN_TYPES = (np.bool_, pd.Int8Dtype())

UINT_TYPES = (np.uint8, np.uint16, np.uint32, np.uint64)

UINT_NULLABLE_TYPES = (
    pd.UInt8Dtype(),
    pd.UInt16Dtype(),
    pd.UInt32Dtype(),
    pd.UInt64Dtype(),
)

INT_TYPES = (np.int8, np.int16, np.int32, np.int64)

INT_NULLABLE_TYPES = (
    pd.Int8Dtype(),
    pd.Int16Dtype(),
    pd.Int32Dtype(),
    pd.Int64Dtype(),
)

FLOAT_TYPES = (np.float16, np.float32, np.float64)



# -*- coding: utf-8 -*-
"""Pandas Downcast
===============

Safely infer minimum viable schema for Pandas DataFrames and Series.

Examples:
    import pdcast as pdc
    import numpy as np
    import pandas as pd

    data = {
        "integers": np.linspace(1, 100, 100),
        "floats": np.linspace(1, 1000, 100).round(2),
        "booleans": np.random.choice([1, 0], 100),
        "categories": np.random.choice(["foo", "bar", "baz"], 100),
    }

    df = pd.DataFrame(data)

    # Downcast DataFrame to minimum viable schema.
    df_downcast = pdc.downcast(df)

    # Infer minimum schema from DataFrame.
    schema = pdc.infer_schema(df)

    # Coerce DataFrame to schema - required if converting float to Pandas Integer.
    df_new = pdc.coerce_df(df, schema)

"""

__author__ = """Dominic Thorn"""
__email__ = "dominic.thorn@gmail.com"
__version__ = "1.2.1"

from pdcast.core import coerce_df, coerce_series, downcast, infer_schema, options



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pdcast']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=0.24']

extras_require = \
{':python_version < "3.7"': ['numpy>=1.16.5', 'dataclasses'],
 ':python_version >= "3.7"': ['numpy>=1.17']}

setup_kwargs = {
    'name': 'pandas-downcast',
    'version': '1.2.1',
    'description': 'Shrink Pandas DataFrames with precision safe schema inference.',
    'long_description': 'Pandas Downcast\n===============\n\n[![image](https://img.shields.io/pypi/v/pandas-downcast.svg)](https://pypi.python.org/pypi/pandas-downcast)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/pandas-downcast.svg)](https://pypi.python.org/pypi/pandas-downcast/)\n[![Build Status](https://travis-ci.com/domvwt/pandas-downcast.svg?branch=main)](https://travis-ci.com/domvwt/pandas-downcast)\n[![codecov](https://codecov.io/gh/domvwt/pandas-downcast/branch/main/graph/badge.svg?token=TQPLURKQ9Z)](https://codecov.io/gh/domvwt/pandas-downcast)\n\nShrink [Pandas](https://pandas.pydata.org/) DataFrames with precision safe schema inference.\n`pandas-downcast` finds the minimum viable type for each column, ensuring that resulting values\nare within tolerance of original values.\n\n## Installation\n\n```bash\npip install pandas-downcast\n```\n\n## Dependencies\n\n* python >= 3.6\n* pandas\n* numpy\n\n## License\n\n[MIT](https://opensource.org/licenses/MIT)\n\n## Usage\n\n```python\nimport pdcast as pdc\nimport numpy as np\nimport pandas as pd\n\ndata = {\n    "integers": np.linspace(1, 100, 100),\n    "floats": np.linspace(1, 1000, 100).round(2),\n    "booleans": np.random.choice([1, 0], 100),\n    "categories": np.random.choice(["foo", "bar", "baz"], 100),\n}\n\ndf = pd.DataFrame(data)\n\n# Downcast DataFrame to minimum viable schema.\ndf_downcast = pdc.downcast(df)\n\n# Infer minimum schema for DataFrame.\nschema = pdc.infer_schema(df)\n\n# Coerce DataFrame to schema - required if converting float to Pandas Integer.\ndf_new = pdc.coerce_df(df, schema)\n```\n\nSmaller data types -> smaller memory footprint.\n\n```python\ndf.info()\n# <class \'pandas.core.frame.DataFrame\'>\n# RangeIndex: 100 entries, 0 to 99\n# Data columns (total 4 columns):\n#  #   Column      Non-Null Count  Dtype  \n# ---  ------      --------------  -----  \n#  0   integers    100 non-null    float64\n#  1   floats      100 non-null    float64\n#  2   booleans    100 non-null    int64  \n#  3   categories  100 non-null    object \n# dtypes: float64(2), int64(1), object(1)\n# memory usage: 3.2+ KB\n\ndf_downcast.info()\n# <class \'pandas.core.frame.DataFrame\'>\n# RangeIndex: 100 entries, 0 to 99\n# Data columns (total 4 columns):\n#  #   Column      Non-Null Count  Dtype   \n# ---  ------      --------------  -----   \n#  0   integers    100 non-null    uint8   \n#  1   floats      100 non-null    float32 \n#  2   booleans    100 non-null    bool    \n#  3   categories  100 non-null    category\n# dtypes: bool(1), category(1), float32(1), uint8(1)\n# memory usage: 932.0 bytes\n```\n\nNumerical data types will be downcast if the resulting values are within tolerance of the original values.\nFor details on tolerance for numeric comparison, see the notes on [`np.allclose`](https://numpy.org/doc/stable/reference/generated/numpy.allclose.html).\n\n```python\nprint(df.head())\n#    integers  floats  booleans categories\n# 0       1.0    1.00         1        foo\n# 1       2.0   11.09         0        baz\n# 2       3.0   21.18         1        bar\n# 3       4.0   31.27         0        bar\n# 4       5.0   41.36         0        foo\n\nprint(df_downcast.head())\n#    integers     floats  booleans categories\n# 0         1   1.000000      True        foo\n# 1         2  11.090000     False        baz\n# 2         3  21.180000      True        bar\n# 3         4  31.270000     False        bar\n# 4         5  41.360001     False        foo\n\n\nprint(pdc.options.ATOL)\n# >>> 1e-08\n\nprint(pdc.options.RTOL)\n# >>> 1e-05\n```\n\nTolerance can be set at module level or passed in function arguments.\n\n```python\npdc.options.ATOL = 1e-10\npdc.options.RTOL = 1e-10\ndf_downcast_new = pdc.downcast(df)\n```\n\nOr\n\n```python\ninfer_dtype_kws = {\n    "ATOL": 1e-10,\n    "RTOL": 1e-10\n}\ndf_downcast_new = pdc.downcast(df, infer_dtype_kws=infer_dtype_kws)\n```\n\nThe `floats` column is now kept as `float64` to meet the tolerance requirement.\nValues in the `integers` column are still safely cast to `uint8`.\n\n```python\ndf_downcast_new.info()\n# <class \'pandas.core.frame.DataFrame\'>\n# RangeIndex: 100 entries, 0 to 99\n# Data columns (total 4 columns):\n#  #   Column      Non-Null Count  Dtype   \n# ---  ------      --------------  -----   \n#  0   integers    100 non-null    uint8   \n#  1   floats      100 non-null    float64 \n#  2   booleans    100 non-null    bool    \n#  3   categories  100 non-null    category\n# dtypes: bool(1), category(1), float64(1), uint8(1)\n# memory usage: 1.3 KB\n```\n\nInferred schemas can be restricted to Numpy data types only.\n\n```python\n# Downcast DataFrame to minimum viable Numpy schema.\ndf_downcast = pdc.downcast(df, numpy_dtypes_only=True)\n\n# Infer minimum  Numpy schema for DataFrame.\nschema = pdc.infer_schema(df, numpy_dtypes_only=True)\n```\n\n## Example\n\nThe following example shows how downcasting data often leads to size reductions of **greater than 70%**, depending on the original types.\n\n```python\nimport pdcast as pdc\nimport pandas as pd\nimport seaborn as sns\n\ndf_dict = {df: sns.load_dataset(df) for df in sns.get_dataset_names()}\n\nresults = []\n\nfor name, df in df_dict.items():\n    size_pre = df.memory_usage(deep=True).sum()\n    df_post = pdc.downcast(df)\n    size_post = df_post.memory_usage(deep=True).sum()\n    shrinkage = int((1 - (size_post / size_pre)) * 100)\n    results.append(\n        {"dataset": name, "size_pre": size_pre, "size_post": size_post, "shrink_pct": shrinkage}\n    )\n\nresults_df = pd.DataFrame(results).sort_values("shrink_pct", ascending=False).reset_index(drop=True)\nprint(results_df)\n```\n\n```\n           dataset  size_pre  size_post  shrink_pct\n0             fmri    213232      14776          93\n1          titanic    321240      28162          91\n2        attention      5888        696          88\n3         penguins     75711       9131          87\n4             dots    122240      17488          85\n5           geyser     21172       3051          85\n6           gammas    500128     108386          78\n7         anagrams      2048        456          77\n8          planets    112663      30168          73\n9         anscombe      3428        964          71\n10            iris     14728       5354          63\n11        exercise      3302       1412          57\n12         flights      3616       1888          47\n13             mpg     75756      43842          42\n14            tips      7969       6261          21\n15        diamonds   3184588    2860948          10\n16  brain_networks   4330642    4330642           0\n17     car_crashes      5993       5993           0\n```\n',
    'author': 'Dominic Thorn',
    'author_email': 'dominic.thorn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/domvwt/pandas-downcast',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""Core functions for downcasting Pandas DataFrames and Series."""

from dataclasses import dataclass
from typing import Any, Dict, Hashable, Iterable, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame
from pandas.core.series import Series

import pdcast.types as tc
from pdcast.types import FrameOrSeries


@dataclass
class Options:
    """Config options for `pdcast`."""

    RTOL = 1e-05
    """Default relative tolerance for numpy inexact numeric comparison
    See: https://numpy.org/doc/stable/reference/generated/numpy.allclose.html"""

    ATOL = 1e-08
    """Default absolute tolerance for numpy inexact numeric comparison
    See: https://numpy.org/doc/stable/reference/generated/numpy.allclose.html"""

    numpy_dtypes_only = False


options = Options()


class _ValidTypeFound(Exception):
    """ """


def infer_dtype(
    series: Series,
    cat_thresh: float = 0.8,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
    numpy_dtypes_only: Optional[bool] = None,
) -> Any:
    """Determine smallest viable type for Pandas Series.

    Args:
        series (Series): Pandas Series.
        cat_thresh (float): Categorical value threshold. (Default value = 0.8)
            Non-numeric variables with proportion of unique values less than
            `cat_thresh` will be cast to Categorical type.
        rtol (float): Absolute tolerance for numeric equality. (Default value = None)
        atol (float): Relative tolerance for numeric equality. (Default value = None)
        numpy_dtypes_only (bool): Use only Numpy dtypes for schema. (Default value = None)

    Returns:
        Smallest viable Numpy or Pandas data type for `series`.

    """

    original_dtype = series.dtype
    valid_type = None

    numpy_dtypes_only = numpy_dtypes_only or options.numpy_dtypes_only

    def assign_valid_type(data_type) -> None:
        nonlocal valid_type
        valid_type = data_type
        raise _ValidTypeFound

    def first_valid_type(srs, candidate_types) -> None:
        for data_type in candidate_types:
            if type_cast_valid(srs, data_type):
                assign_valid_type(data_type)

    try:
        if pd.api.types.is_numeric_dtype(original_dtype):
            val_range = series.min(), series.max()
            val_min, val_max = val_range[0], val_range[1]
            is_signed = val_range[0] < 0
            is_nullable = series.isna().any()

            srs_mod_one = np.mod(series.fillna(0), 1)
            is_decimal = not all(close_to_val(srs_mod_one, 0))

            if not is_decimal:
                # Identify 1 / 0 flags and cast to `np.bool`
                if (
                    close_to_0_or_1(val_min, rtol, atol)
                    and close_to_0_or_1(val_max, rtol, atol)
                    and series.dropna().unique().shape[0] == 2
                ):
                    # Convert values close to zero to exactly zero and values close to one to
                    # exactly one so that Pandas allows recast to int type
                    series = np.where(close_to_val(series, 0, rtol, atol), 0, series)
                    series = np.where(close_to_val(series, 1, rtol, atol), 1, series)
                    series = Series(series)

                    first_valid_type(series, tc.BOOLEAN_TYPES)

                if is_nullable and not numpy_dtypes_only:
                    if is_signed:
                        first_valid_type(series, tc.INT_NULLABLE_TYPES)

                    # Unsigned
                    else:
                        # Other integer types
                        first_valid_type(series, tc.UINT_NULLABLE_TYPES)

                # Not nullable
                elif not is_nullable:
                    if is_signed:
                        first_valid_type(series, tc.INT_TYPES)
                    # Unsigned
                    else:
                        first_valid_type(series, tc.UINT_TYPES)
            # Float type
            first_valid_type(series, tc.FLOAT_TYPES)

        elif series.isna().all():
            assign_valid_type(tc.ALL_NAN_TYPE)

        # Non-numeric (and not a date)
        elif original_dtype == object:
            if not numpy_dtypes_only:
                unique_count = series.dropna().unique().shape[0]
                srs_length = series.dropna().shape[0]
                unique_pct = unique_count / srs_length
                # Cast to `categorical` if percentage of uniques value less than threshold
                if unique_pct < cat_thresh:
                    assign_valid_type(pd.CategoricalDtype())
            assign_valid_type(np.object_)

    except _ValidTypeFound:
        return valid_type

    # Keep original type if nothing else works
    except TypeError:
        pass

    return original_dtype


def infer_schema(
    data: FrameOrSeries,
    include: Optional[Iterable[Hashable]] = None,
    exclude: Optional[Iterable[Hashable]] = None,
    sample_size: int = 10_000,
    numpy_dtypes_only: Optional[bool] = None,
    infer_dtype_kws: Optional[Dict[str, Any]] = None,
) -> Dict[Any, Any]:
    """Infer minimum viable schema for `data`.

    Args:
        data (FrameOrSeries): Pandas DataFrame or Series.
        include (Iterable[Hashable]): Columns to include. (Default value = None)
            Excludes all other columns if defined.
        exclude (Iterable[Hashable]): Columns to exclude. (Default value = None)
        sample_size (int): Number of records to take from head and tail. (Default value = 10_000)
        numpy_dtypes_only (bool): Use only Numpy dtypes for schema. (Default value = None)
        infer_dtype_kws (Dict[Any, Any]): Keyword arguments for `infer_dtype`. (Default value = None)

    Returns:
        Dict[str, Any]: Inferred schema.

    """
    if not isinstance(data, (DataFrame, Series)):
        raise TypeError(type(data))
    data = data.copy()
    infer_dtype_kws = infer_dtype_kws or {}
    # Use head and tail in case data is sorted
    if sample_size and data.shape[0] > sample_size:
        data = take_head_and_tail(data)
    if isinstance(data, Series):
        schema = {
            data.name: infer_dtype(
                data, numpy_dtypes_only=numpy_dtypes_only, **infer_dtype_kws
            )
        }
    else:  # DataFrame
        target_cols = include or data.columns
        if exclude:
            target_cols = [col for col in target_cols if col not in set(exclude)]
        schema = {
            col: (
                infer_dtype(srs, numpy_dtypes_only=numpy_dtypes_only, **infer_dtype_kws)
                if col in set(target_cols)
                else srs.dtype
            )
            for col, srs in data.iteritems()
        }
    return schema


def coerce_df(df: DataFrame, schema: Dict[Hashable, Any]) -> FrameOrSeries:
    """Coerce DataFrame to `schema`.

    Args:
        df (DataFrame): Pandas DataFrame.
        schema (dict): Target schema.

    Returns:
        DataFrame: Pandas DataFrame with `schema`.

    """
    df = df.copy()
    try:
        df = df.astype(schema)
    except TypeError:
        for col, dtype in schema.items():
            df[col] = coerce_series(df[col], dtype)
    return df


def coerce_series(series: Series, dtype: Any) -> FrameOrSeries:
    """Coerce Series to `dtype`.

    Args:
        series (Series): Pandas Series.
        dtype (Any): Target data type.

    Returns:
        Series: Pandas Series of type `dtype`.

    """
    series = series.copy()
    try:
        series = series.astype(dtype)
    except TypeError:
        # TypeError thrown when converting float to int without rounding
        series = series.round(0).astype(dtype)
    return series


def downcast(
    data: FrameOrSeries,
    include: Optional[Iterable[Hashable]] = None,
    exclude: Optional[Iterable[Hashable]] = None,
    return_schema: bool = False,
    sample_size: int = 10_000,
    numpy_dtypes_only: bool = False,
    infer_dtype_kws: Optional[Dict[str, Any]] = None,
) -> Union[DataFrame, Series, Tuple[FrameOrSeries, Dict[Hashable, Any]]]:
    """Infer and apply minimum viable schema.

    Args:
        data (FrameOrSeries):
        include (Iterable[Hashable]): Columns to include. (Default value = None)
            Excludes all other columns if defined.
        exclude (Iterable[Hashable]): Columns to exclude. (Default value = None)
        return_schema (bool): Return inferred schema if True. (Default value = False)
        sample_size (int): Number of records to take from head and tail. (Default value = 10_000)
        numpy_dtypes_only (bool): Use only Numpy dtypes for schema. (Default value = None)
        infer_dtype_kws (Dict[str, Any]): Keyword arguments for `infer_dtype`. (Default value = None)

    Returns:
        FrameOrSeries: Downcast Pandas DataFrame or Series.
        Dict[Any, Any]: Inferred schema. (if `return_schema` is True)

    """
    data = data.copy()
    schema = infer_schema(
        data,
        include=include,
        exclude=exclude,
        sample_size=sample_size,
        numpy_dtypes_only=numpy_dtypes_only,
        infer_dtype_kws=infer_dtype_kws,
    )
    if isinstance(data, Series):
        dtype = schema[data.name]
        data = coerce_series(data, dtype)
    else:  # DataFrame
        data = coerce_df(data, schema)
    if return_schema:
        return data, schema
    return data


def take_head_and_tail(data: FrameOrSeries, sample_size: int = 10_000) -> FrameOrSeries:
    """Take head and tail of DataFrame or Series.

    Args:
        data (FrameOrSeries): Pandas DataFrame or Series.
        sample_size (int): Number of records to take. (Default value = 10_000)

    Returns:
        FrameOrSeries: Resampled `data`.

    """
    if data.shape[0] > sample_size:
        half_sample = sample_size // 2
        data = pd.concat([data[:half_sample], data[-half_sample:]])
    return data


def type_cast_valid(
    series: Series,
    data_type: Any,
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
) -> bool:
    """Check `series` can be cast to `data_type` without loss of information.

    Args:
        series (Series): Pandas Series.
        data_type (Any): Any data type.
        rtol (float): Absolute tolerance for numeric equality. (Default value = None)
        atol (float): Relative tolerance for numeric equality. (Default value = None)

    Returns:
        bool: True if type if valid, False otherwise.

    """
    try:
        srs_new = series.astype(data_type)
    except TypeError:
        return False
    sdtype = series.dtype
    rtol = rtol or options.RTOL
    atol = atol or options.ATOL
    if is_numeric_typelike(sdtype) and is_numeric_typelike(data_type):
        return np.allclose(srs_new, series, equal_nan=True, rtol=rtol, atol=atol)
    return bool(Series(series == srs_new).all())


def is_numeric_typelike(dtype) -> bool:
    """Check whether `dtype` is a numeric type or class.

    Args:
        dtype: Any data type class or object.

    Returns:
        bool: True if numeric type, False otherwise.

    """
    return (isinstance(dtype, type) and issubclass(dtype, np.number)) or (
        isinstance(dtype, np.dtype) and np.issubdtype(dtype, np.number)
    )


def close_to_val(
    series: Series,
    val: Union[int, float],
    rtol: Optional[float] = None,
    atol: Optional[float] = None,
) -> Series:
    """Check all `series` values close to `val`.

    Args:
        series (Series): Pandas Series.
        val (Union[int, float]): Value for comparison.
        rtol (float): Absolute tolerance for numeric equality. (Default value = None)
        atol (float): Relative tolerance for numeric equality. (Default value = None)

    Returns:
        bool: True if all close to `val`, False otherwise.

    """
    rtol = rtol or options.RTOL
    atol = atol or options.ATOL
    try:
        return np.isclose(series, val, rtol=rtol, atol=atol)
    except TypeError:
        return pd.Series(series == val)


def close_to_0_or_1(
    num: np.number, rtol: Optional[float] = None, atol: Optional[float] = None
) -> bool:
    """Check if `num` is close to zero or one.

    Args:
        num (np.number): Number for comparison.
        rtol (float): Absolute tolerance for numeric equality. (Default value = None)
        atol (float): Relative tolerance for numeric equality. (Default value = None)

    Returns:
        bool: True if all close to 0 or 1, False otherwise.

    """
    rtol = rtol or options.RTOL
    atol = atol or options.ATOL
    return bool(
        np.isclose(num, 0, rtol=rtol, atol=atol)
        or np.isclose(num, 1, rtol=rtol, atol=atol)
    )
